import requests
from json import loads, dumps
from urllib.parse import quote
from html import escape
from re import sub

class RequestError(Exception):
    pass
class ResponseError(Exception):
    pass
class AuthError(Exception):
    pass
class PathError(Exception):
    pass

class Item:
    path:  str
    name:  str
    icon:  str
    id:    str
    owner: str = None
    def __init__(self, path: str, name: str, icon: str, id: str, owner: str = None) -> None:
        self.path  = path
        self.name  = name
        self.icon  = icon
        self.id    = id
        self.owner = owner

class File(Item):
    size: str = None
    def __init__(self, path: str, name: str, icon: str, id: str, owner: str = None, size: str = None) -> None:
        self.path  = path
        self.name  = name
        self.icon  = icon
        self.id    = id
        self.owner = owner
        self.size  = size

class Folder(Item):
    pass

class BakalariFileExplorer:
    url = ""
    session = None
    authenticated = False
    idDict = {}
    idRoot = ""

    def __init__(self, url: str, idRoot: str = "a:"):
        self.url = url
        self.idRoot = idRoot
        self.session = requests.session()

    # Define headers
    headers = {
        "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
    }

    def authWithCredentials(self, username: str, password: str) -> None:
        """Authenticate with credentials"""
        # Prepare payload for our auth request
        payload = ""
        payload += "&username=" + username
        payload += "&password=" + password
        payload += "&persistent=true"
        payload += "&returnUrl="
        payload += "&login="

        # Prepare url for our request
        url = self.url + "/Login"

        # Try to authenticate with our credentials
        res = self.session.post(url, headers = self.headers, data = payload)
        if(res.status_code != 200):
            raise RequestError(res.content)
        
        if(not "BakaAuth" in self.session.cookies):
            raise AuthError("Wrong credentials.")
        self.authenticated = True
    
    def authWithCookie(self, BakaAuth: str) -> None:
        """Authenticate with cookie"""
        # Set cookie to this session
        self.session.cookies["BakaAuth"] = BakaAuth

        # Prepare url for our test request
        url = self.url + "/Login"

        # Try to authenticate with our cookie
        res = self.session.post(url, headers = self.headers)
        if(res.status_code != 200):
            raise RequestError(res.content)
        
        # This check will not work
        # TODO: implement better check
        if(not "BakaAuth" in self.session.cookies):
            raise AuthError("Uneatable cookie.")
        self.authenticated = True
    
    def getAuthCookie(self) -> str:
        if(not self.authenticated): raise AuthError("You must authenticate before using this method.")
        return self.session.cookies["BakaAuth"]
    
    def downloadFile(self, id: str, localPath: str) -> None:
        if(not self.authenticated): raise AuthError("You must authenticate before using this method.")
        url = self.url + "/next/getDocument.aspx?id=" + id.split(":")[-1]
        res = self.session.get(url)
        with open(localPath, 'wb') as file:
            file.write(res.content)

    def readFile(self, id: str) -> bytes:
        if(not self.authenticated): raise AuthError("You must authenticate before using this method.")
        url = self.url + "/next/getDocument.aspx?id=" + id.split(":")[-1]
        res = self.session.get(url)
        return res.content

    def listFolder(self, path: str) -> list:
        if(not self.authenticated): raise AuthError("You must authenticate before using this method.")
        idPath = self._generateIdPath(path)
        items = self._listFolder(path, idPath)
        
        # Generate list of Item-s from JSON list
        ret = []
        for item in items:
            if(item['it'] == "Folder"): ret.append(Folder(
                path  = path,
                name  = item['n'],
                icon  = self.url + "/next/" + item['i'],
                id    = item['id'],
                owner = None if(not item['md']) else item['md']['Owner']
            ))
            elif(item['it'] == "File"): ret.append(File(
                path = path,
                name = item['n'],
                icon = self.url + "/next/" + item['i'],
                id    = item['id'],
                owner = None if(not item['md']) else item['md']['Owner'],

                size = None if(not item['md']) else item['md']['Size']
            ))
            else:
                print("[listFolder]", "Unsupported type:", item['it'])
        return ret

    def _generateIdPath(self, path: str):
        currentIdPath = []
        currentPath = ""
        if(path != ""):
            for folder in path.split("\\"):
                if(not currentPath + "\\" + folder in self.idDict):
                    # We must get id of this folder
                    folderItem = None
                    for item in self._listFolder(currentPath, currentIdPath):
                        if(item['n'] == folder):
                            folderItem = item
                            break
                    if(not folderItem): raise PathError("Directory '{}' in '{}' does not exist.".format(
                        folder,
                        currentPath
                    ))
                    self.idDict[currentPath + "\\" + folder] = item['id']
                
                currentIdPath.append(self.idDict[currentPath + "\\" + folder])
                currentPath += ("" if(len(currentPath) == 0) else "\\") + folder
        return currentIdPath

    def _listFolder(self, path: str, idPath: list) -> list:
        # Prepare payload for our request
        payload = ""
        payload += "&__VIEWSTATE="
        #payload += "&hfViewMode=ShowThumbs"
        payloadFileManager = dumps(
            {
                "currentPath": path + "|false",
                "currentIdPath": [self.idRoot] + idPath
            }, ensure_ascii=False)
        payload += "&ctl00%24cphmain%24FileManager=" + quote(escape(payloadFileManager))
        payload += "&__CALLBACKID=ctl00%24cphmain%24FileManager"
        # I'm not really sure what this last parameter means but it looks
        # like it's working in all directories and it does not work
        # without it so i'm leaving it here
        payload += "&__CALLBACKPARAM=c0%3AGETFILELIST%7C%7B%22name%22%3Anull%7D"

        # Prepare url for our request
        url = self.url + "/next/dokumentyPrehled.aspx"

        # Post request to the server
        res = self.session.post(url, headers = self.headers, data = payload)
        if(res.status_code != 200):
            raise RequestError(res.content)
        responseJs = res.content.decode();

        # Extract items list from the plaintext
        # I bet this could be done in one line by using regex but idc
        itemsKeyword = "'items':"
        if(not itemsKeyword in responseJs):
            raise ResponseError("No items found for this path.")
        index = responseJs.index(itemsKeyword) + len(itemsKeyword)
        content = ""
        escaped = 0
        while index < len(responseJs):
            if(not escaped):
                # Change ' to " or just copy the character
                if(responseJs[index] == "'"):
                    content += '"'
                else:
                    content += responseJs[index]

                # Detect escaping
                if(responseJs[index] == "\\"):
                    escaped = True
                # Detect end of items list
                elif(responseJs[index] == "]"):
                    break
            else:
                # Just copy the character
                content += responseJs[index]
                escaped = False
            index += 1

        # Patch the string because it contains some JS scripts that
        # JSON parser would not understand
        contentPatched = sub(",\"\w*\":new Date\([0-9,]*\)", "", content)
        # Parse it and return
        return loads(contentPatched)